/*
    Copyright (c) 1997 Eliot W. Dudley.  All rights reserved.

       edudley@servtech.com
       XIDAR
       3388 STATE RT 370
       CATO NY  13033

*/
/*
       1-Wire, DS, DS1820, DS2405, DS2407, DS9097, and MicroLan
          are trademarks and/or registered trademarks of
                 Dallas Semiconductor Coporation
*/
//=============================================================================
  #include <stdio.h>
  #include <stdarg.h>
  #include <stdlib.h>
//#include <memory.h>
//#include <string.h>

  #include <time.h>
  #include <fcntl.h>
  #include <setjmp.h>

  #include <unistd.h>
  #include <sys/time.h>
  #include <sys/types.h>
  #include <asm/bitops.h>

  #include <termios.h>
  #include <sys/ioctl.h>
  #include <linux/serial.h>
  #include <linux/tty.h>

  #include "util.h"
  #include "b.h"

  #include "mlfnc.h"
  #include "ml.h"
  #include "mlerr.h"
  #include "mld.h"
  #include "mlbus.h"

typedef struct serial_struct      SERIAL_STRUCT;
typedef struct termios            TERMIOS;

struct MLPORTt {
    /// Serial port device name.
   char                   *name;
    /// File descriptor for the serial port device.
   int                     fd;
    /// For 115.2Kbps and custom baud rate control.
   SERIAL_STRUCT           serial_struct;
    /// Normal termios line control.
   TERMIOS                 termios;
    /// Buf for expanding i/o into one 115.2Kbps byte per bit of information.
   BYTE                    buf[8 * ML_MAX_IO];
};

//</XMP><A NAME="MLOpen"><XMP>
/*/
   Open the serial device and store the resulting file descriptor.  Some
    of the usual ioctl/termios/tcsetattr stuff, some not-so-usual 16550A
    custom speed stuff.
    Source: \URL{../mld.c.html#MLOpen}
*/
int
MLOpen(
   pML           pml,
   char         *dev
) {
   pMLPORT       pmlport;
   STACK_CHECK();
   ENTER(MLOpen);

   if (!dev) {
      Fatal(NO_SERIAL_DEVICE_SPECIFIED, "");
   }
   MALLOC(pmlport, 1, MLPORT);
   pml->pmlport = pmlport;
   MALLOC(pmlport->name, strlen(dev) + 1, char);
   if (0 > (pmlport->fd =
    open(strcpy(pmlport->name, dev), O_RDWR))) {
      Fatal(OPEN_FAILED, FATAL_LINE("%s %s"), pmlport->name, strerror(errno));
   }
   if (0 > ioctl(pmlport->fd, TIOCGSERIAL, &pmlport->serial_struct)) {
      Fatal(IOCTL_FAILED, FATAL_LINE("%s %s"), pmlport->name, strerror(errno));
   }
   if (0 > tcgetattr(pmlport->fd, &pmlport->termios)) {
      Fatal(TCGETATTR_FAILED,
       FATAL_LINE("%s %s"), pmlport->name, strerror(errno));
      return(0);
   }
   tcflush(pmlport->fd, TCIOFLUSH);
   pmlport->serial_struct.type = PORT_16550A;
   pmlport->serial_struct.flags =
    (pmlport->serial_struct.flags & ~ASYNC_SPD_MASK) | ASYNC_SPD_CUST;
   pmlport->serial_struct.custom_divisor = 1;
   ioctl(pmlport->fd, TIOCSSERIAL, &pmlport->serial_struct);

   cfsetospeed(&pmlport->termios, B38400);
   cfsetispeed(&pmlport->termios, B38400);
   pmlport->termios.c_oflag = 0;
   pmlport->termios.c_lflag = 0;
   tcsetattr(pmlport->fd, TCSAFLUSH, &pmlport->termios);

   POP_CALL_TRACE();
   return(1);
}

//</XMP><A NAME="MLWrite"><XMP>
/*/
   Write count bits from buf to the bus object pml.  Read the state of
    the bus for each bit written and record in the same buf.  According
    to flag options, optionally reset the bus before the write, and/or
    timestamp the pmlnode passed by reference.  Returns error code less
    than zero on error, else the number of bits written.
    Source: \URL{../mld.c.html#MLWrite}
*/
int
MLWrite_(
   pML                     pml,
   char                   *buf,
   int                     count,
   int                     flag,
   pMLNODE                 pmlnode
) {
   int                     ret;
   int                     i;
   BYTE                    w;
   BYTE                    r;
   struct timeval          tv;
   pMLPORT                 pmlport;

   pmlport = pml->pmlport;

   //==========================================================================
   if ((ML_RESET_ADDR & flag) || (ML_RESET_POWER & flag)) {
      //=Go to eight bit. 10.47KBaud for reset addr, 450 for reset power.======
      tcflush(pmlport->fd, TCIOFLUSH);
      pmlport->serial_struct.custom_divisor = 
       (ML_RESET_ADDR & flag) ? 11 : 256;
      ioctl(pmlport->fd, TIOCSSERIAL, &pmlport->serial_struct);
      pmlport->termios.c_cflag = 0
       // | CS6            // 0000020
       // | CS7            // 0000040
          | CS8            // 0000060
       // | CSTOPB         // 0000100
          | CREAD          // 0000200
       // | PARENB         // 0000400
       // | PARODD         // 0001000
          | HUPCL          // 0002000
          | CLOCAL         // 0004000
          | B38400         // 0000017
      ;
      tcsetattr(pmlport->fd, TCSANOW, &pmlport->termios);
      //=======================================================================

      //=Send the reset pulse.=================================================
      // (115200Hz clock)/(Divisor = 11) --> 95.5uSec per bit.
      // LSB first, start bit is low, stop bit is high.
      // So five bits low, five bits high is 477uSec reset pulse.

      // (115200Hz clock)/(Divisor = 256) --> 2222.2 uSec per bit.
      // LSB first, start bit is low, stop bit is high.
      // So nine bits low, one bit high is 20ms power reset pulse.

      w = (ML_RESET_ADDR & flag) ? 0xf0 : 0x00;
      if (0 > write(pmlport->fd, &w, 1)) {
         Fatal(WRITE_FAILED, 
          FATAL_LINE("%s %s"), pmlport->name, strerror(errno));
      }
      //=======================================================================
      //=Convert bits to byte stream while waiting.============================
      for (i = 0; i < count; ++i) {
         pmlport->buf[i] = "\x00\xff"[test_bit(i, buf)];
      }
      //=======================================================================

      //=Wait at least 2ms/30ms for response.==================================
      if ((ret =
       ReadYield(pmlport->fd, (ML_RESET_ADDR & flag) ? 2000 : 30000))) {
         if (0 > read(pmlport->fd, &r, 1)) {
            Fatal(READ_FAILED, 
             FATAL_LINE("%s %s"), pmlport->name, strerror(errno));
         }
         if (ML_RESET_ADDR & flag) {
            //=No reset pulse detected, we're talking to the bus but no answer.
            if (r == w) {
               return(-NO_DEVICES_FOUND);
            }
         } else {
            return(1);
         }
      } else {
         //=We're not getting echo on the bus.=================================
         return(-XIDARML_BUS_FAILURE);
      }
      //=======================================================================
      // Go to 115.2KBaud, 6 bit.==============================================
      pmlport->serial_struct.custom_divisor = 1;
      ioctl(pmlport->fd, TIOCSSERIAL, &pmlport->serial_struct);
      pmlport->termios.c_cflag = 0
          | CS6            // 0000020
       // | CS7            // 0000040
       // | CS8            // 0000060
       // | CSTOPB         // 0000100
          | CREAD          // 0000200
       // | PARENB         // 0000400
       // | PARODD         // 0001000
          | HUPCL          // 0002000
          | CLOCAL         // 0004000
          | B38400         // 0000017
      ;
      tcsetattr(pmlport->fd, TCSANOW, &pmlport->termios);
      //=======================================================================

   } else {

      //=Convert bits to byte stream for write.================================
      for (i = 0; i < count; ++i) {
         pmlport->buf[i] = "\x00\xff"[test_bit(i, buf)];
      }
      //=======================================================================
   }
   //==========================================================================

   //==========================================================================
   if (0 > write(pmlport->fd, pmlport->buf, count)) {
      Fatal(WRITE_FAILED, FATAL_LINE("%s %s"), pmlport->name, strerror(errno));
   }
   //==========================================================================

   //==========================================================================
   if ((pmlnode) && (ML_TIMESTAMP & flag)) {
      gettimeofday(&tv, (struct timezone *)0);
      pmlnode->tv = tv;
   }
   //==========================================================================

   //=Wait for response.  Read them back.======================================
   memset(buf, 0, (count + 7) / 8);

   for (i = 0;;) {
      if ((ret = ReadYield(pmlport->fd, 70 * (count - i)))) {
         if (0 > (ret = read(pmlport->fd, pmlport->buf + i, count - i))) {
            Fatal(READ_FAILED, 
             FATAL_LINE("%s %s"), pmlport->name, strerror(errno));
         }
         if ((i += ret) < count) {
            continue;
         }
         for (i = 0; i < count; ++i) {
            if (b_00000001 & pmlport->buf[i]) {
               set_bit(i, buf);
            }
         }
         return(count);
      }
      break;
   }
   //==========================================================================

   //=We're not getting echo on the bus.=======================================
   return(-XIDARML_BUS_FAILURE);
   //==========================================================================
}

int
MLWrite(
   pML                     pml,
   char                   *buf,
   int                     count,
   int                     flag,
   pMLNODE                 pmlnode
) {
   int                     ret;
   int                     i;
   char                    cuf[1];
   int                     timestamp;

   if (!(pml->mlslomo)) {
      return(MLWrite_(pml, buf, count, flag, pmlnode));
   }
   if (!(count)) {
      return(MLWrite_(pml, buf, count, flag, pmlnode));
   }

   timestamp = (ML_TIMESTAMP & flag);
   flag &= ~ML_TIMESTAMP;

   for (i = 0; i < count; ++i) {
      if (i) {
         flag &= ~(ML_RESET_ADDR | ML_RESET_POWER);
      }
      if ((timestamp) && (i == (count - 1))) {
         flag = ML_TIMESTAMP;
      }
      *cuf = !!(test_bit(i, buf));
      if (0 > (ret = MLWrite_(pml, cuf, 1, flag, pmlnode))) {
         return(ret);
      }
      if (*cuf) {
         set_bit(i, buf);
      } else {
         clear_bit(i, buf);
      }
   }
   return(count);
}
